#ifndef _LINUX_TIVOCONFIG_INIT_H
#define _LINUX_TIVOCONFIG_INIT_H

/*
 * Initialization array for the TiVo configuration registry
 *
 * Copyright (C) 2009 TiVo Inc. All Rights Reserved.
 */

typedef struct {
	TivoConfigSelector selector;
	TivoConfigValue value;
} TivoConfig;


/* Gen07 */
static TivoConfig gGen07ConfigTable[] __initdata = {
    { kTivoConfigBoardID, kTivoConfigBoardIDNeutron },
    { kTivoConfigMemSize, 0x1ffff000 }, // 512 MB, minus one page WAR
    { kTivoConfigContigmemRegion1, 0x00000000 }, /* 0MB - input section now in contigmem8 */
    { kTivoConfigContigmemRegion2, 0x00200000 }, /* 2.0MB */
    { kTivoConfigContigmemRegion3, 0x00527000 }, /* 5280kB */
    { kTivoConfigContigmemRegion8, 0x0B000000 }, /*176.0MB */
    { 0,0 }
};

/* Gen07Leo */
static TivoConfig gLeoConfigTable[] __initdata = {
    { kTivoConfigBoardID, kTivoConfigBoardIDLeo },
    { kTivoConfigMemSize, 0x1ffff000 }, // 512 MB, minus one page WAR
    { kTivoConfigContigmemRegion1, 0x00000000 }, /* 0MB - input section now in contigmem8 */
    { kTivoConfigContigmemRegion2, 0x00200000 }, /* 2.0MB */
    { kTivoConfigContigmemRegion3, 0x00527000 }, /* 5280kB */
    { kTivoConfigContigmemRegion8, 0x08D00000 }, /* 141.0MB */
    { 0,0 }
};

/* Gen07Cerberus */
static TivoConfig gCerberusConfigTable[] __initdata = {
    { kTivoConfigBoardID, kTivoConfigBoardIDCerberus },
    { kTivoConfigMemSize, 0x1ffff000 }, // 512 MB, minus one page WAR
    { kTivoConfigContigmemRegion1, 0x00000000 }, /* 0MB - input section now in contigmem8 */
    { kTivoConfigContigmemRegion2, 0x00200000 }, /* 2.0MB */
    { kTivoConfigContigmemRegion3, 0x00527000 }, /* 5280kB */
    { kTivoConfigContigmemRegion8, 0x09600000 }, /* 150.0MB */
    { 0,0 }
};

/* Gen07Cyclops */
static TivoConfig gCyclopsConfigTable[] __initdata = {
    { kTivoConfigBoardID, kTivoConfigBoardIDCyclops },
    { kTivoConfigMemSize, 0x1ffff000 }, // 512 MB, minus one page WAR
    { kTivoConfigContigmemRegion1, 0x00000000 }, /* 0MB - input section now in contigmem8 */
    { kTivoConfigContigmemRegion2, 0x00200000 }, /* 2.0MB */
    { kTivoConfigContigmemRegion3, 0x00527000 }, /* 5280kB */
    { kTivoConfigContigmemRegion8, 0x08D00000 }, /* 150.0MB */
    { 0,0 }
};

/* Mojave/Thompson HR22 */
static TivoConfig gMojaveConfigTable[] __initdata = {
    { kTivoConfigBoardID, kTivoConfigBoardIDMojave },
    { kTivoConfigMemSize, 0x0ffff000 }, // 256 MB, minus one page WAR
    { kTivoConfigContigmemRegion1, 0x01800000 }, /*24.0MB */
    { kTivoConfigContigmemRegion2, 0x00200000 }, /* 2.0MB */
    { kTivoConfigContigmemRegion3, 0x00527000 }, /* 5280kB */
    { kTivoConfigContigmemRegion8, 0x06000000 }, /*96.0MB */
    { 0,0 }
};

static TivoConfig *gGlobalConfigTable[] __initdata = { 
    gGen07ConfigTable,
    gMojaveConfigTable,
    gLeoConfigTable,
    gCerberusConfigTable,
    gCyclopsConfigTable,
    0
};

#else

#error "Nobody should ever #include <linux/tivoconfig-init.h> more than once!"

#endif /* _LINUX_TIVOCONFIG_INIT_H */
